@@ -167,6 +167,17 @@ static void test_add32(abts_case *tc, void *data)
     ABTS_INT_EQUAL(tc, 27, y32);
 }
 
+static void test_add32_neg(abts_case *tc, void *data)
+{
+    apr_uint32_t oldval;
+    apr_uint32_t y32;
+
+    apr_atomic_set32(&y32, 23);
+    oldval = apr_atomic_add32(&y32, -10);
+    ABTS_INT_EQUAL(tc, 23, oldval);
+    ABTS_INT_EQUAL(tc, 13, y32);
+}
+
 static void test_inc32(abts_case *tc, void *data)
 {
     apr_uint32_t oldval;
@@ -509,6 +520,7 @@ abts_suite *testatomic(abts_suite *suite)
     abts_run_test(suite, test_casptr_equal_nonnull, NULL);
     abts_run_test(suite, test_casptr_notequal, NULL);
     abts_run_test(suite, test_add32, NULL);
+    abts_run_test(suite, test_add32_neg, NULL);
     abts_run_test(suite, test_inc32, NULL);
     abts_run_test(suite, test_set_add_inc_sub, NULL);
     abts_run_test(suite, test_wrap_zero, NULL);
